/* Copyright (C) 2015 Arnaud Bonatti <arnaud.bonatti@gmail.com>
 *
 * This file is part of a GNOME game.
 *
 * This application is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This application is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this application. If not, see <http://www.gnu.org/licenses/>.
 */

/* Labels' tweaks */
label.bold-label {
  font-weight: bold;
}
label.italic-label {
  font-style: italic;
}
grid#scoreboard label {
  padding: 4px 7px 4px 0;
}

/* Themes' dialog listbox */
list#themes-listbox image {
  transition-duration: 200ms;
  opacity: 0;
}

list#themes-listbox row:selected * {
  opacity: 1;
  color:@theme_selected_fg_color;
}
list#themes-listbox row:selected image:backdrop {
  color:@theme_unfocused_selected_fg_color;
}

/* make all widgets change font-size at same time */
label.bold-label,
button,
headerbar,
popover modelbutton {
  transition:all       0    ease,
             font-size 0.3s ease,
             padding   0.3s ease,
             margin    0.3s ease;
}

/* bug #746210: HighContrast doesn't update correctly on ListBoxRow selection
   hicolor is preserved because of a "*:selected *:link:visited" declaration */
#fake-example *:selected *:visited { opacity: 1; }

/*\
* * the end
\*/
